@@ -31,18 +31,31 @@
  */
 
 #include <config.h>
-#include <sys/socket.h>
 #include <sys/types.h>
+#include <glib.h>
+#ifndef G_OS_WIN32
+#include <sys/socket.h>
 #include <sys/utsname.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
+#else
+#define _WIN32_WINNT 0x502
+#include <winsock2.h>
+#include <ws2tcpip.h>
+#include <iphlpapi.h>
+typedef int socklen_t;
+/* from the return value of inet_addr */
+typedef unsigned long in_addr_t;
+#endif
 #include <string.h>
 #include <stdio.h>
 #include <errno.h>
 #include <unistd.h>
+#ifndef G_OS_WIN32
 #include <arpa/inet.h>
 #include <net/if.h>
 #include <ifaddrs.h>
+#endif
 #include <libsoup/soup-headers.h>
 
 #include "gssdp-client.h"
@@ -52,6 +65,10 @@
 #include "gssdp-marshal.h"
 #include "gssdp-protocol.h"
 
+#ifndef INET6_ADDRSTRLEN
+#define INET6_ADDRSTRLEN 46
+#endif
+
 /* Size of the buffer used for reading from the socket */
 #define BUF_SIZE 1024
 
@@ -63,7 +80,7 @@ struct _GSSDPClientPrivate {
         GMainContext      *main_context;
 
         char              *server_id;
-        char              *interface;
+        char              *iface;
         char              *host_ip;
 
         GError            **error;
@@ -98,9 +115,13 @@ gssdp_client_set_main_context (GSSDPClient  *client,
 static char *
 make_server_id                (void);
 static gboolean
-request_socket_source_cb      (gpointer      user_data);
+request_socket_source_cb      (GIOChannel   *source,
+                               GIOCondition  condition,
+                               gpointer      user_data);
 static gboolean
-multicast_socket_source_cb    (gpointer      user_data);
+multicast_socket_source_cb    (GIOChannel   *source,
+                               GIOCondition  condition,
+                               gpointer      user_data);
 static gboolean
 init_network_info             (GSSDPClient  *client);
 
@@ -122,6 +143,20 @@ static void
 gssdp_client_constructed (GObject *object)
 {
         GSSDPClient *client = GSSDP_CLIENT (object);
+        GError *error = NULL;
+#ifdef G_OS_WIN32
+        WSADATA wsaData = {0};
+        if (WSAStartup (MAKEWORD (2,2), &wsaData) != 0) {
+                gchar *message;
+
+                message = g_win32_error_message (WSAGetLastError ());
+                g_set_error_literal (client->priv->error,
+                                     GSSDP_ERROR,
+                                     GSSDP_ERROR_FAILED,
+                                     message);
+                g_free (message);
+        }
+#endif
 
         /* Make sure all network info is available to us */
         if (!init_network_info (client))
@@ -130,46 +165,41 @@ gssdp_client_constructed (GObject *object)
         /* Set up sockets (Will set errno if it failed) */
         client->priv->request_socket =
                 gssdp_socket_source_new (GSSDP_SOCKET_SOURCE_TYPE_REQUEST,
-                                         gssdp_client_get_host_ip (client));
+                                         gssdp_client_get_host_ip (client),
+                                         &error);
         if (client->priv->request_socket != NULL) {
-                g_source_set_callback
-                        ((GSource *) client->priv->request_socket,
-                         request_socket_source_cb,
-                         client,
-                         NULL);
+                gssdp_socket_source_set_callback
+                        (client->priv->request_socket,
+                        (GSourceFunc) request_socket_source_cb,
+                        client);
         } else {
                 goto errors;
         }
 
         client->priv->multicast_socket =
                 gssdp_socket_source_new (GSSDP_SOCKET_SOURCE_TYPE_MULTICAST,
-                                         gssdp_client_get_host_ip (client));
+                                         gssdp_client_get_host_ip (client),
+                                         &error);
         if (client->priv->multicast_socket != NULL) {
-                g_source_set_callback
-                        ((GSource *) client->priv->multicast_socket,
-                         multicast_socket_source_cb,
-                         client,
-                         NULL);
+                gssdp_socket_source_set_callback
+                        (client->priv->multicast_socket,
+                         (GSourceFunc) multicast_socket_source_cb,
+                         client);
         }
 
  errors:
         if (!client->priv->request_socket || !client->priv->multicast_socket) {
-                if (client->priv->error)
-                        g_set_error_literal (client->priv->error,
-                                             GSSDP_ERROR,
-                                             GSSDP_ERROR_FAILED,
-                                             g_strerror (errno));
-
+                if (client->priv->error) {
+                        g_propagate_error (client->priv->error, error);
+                }
                 return;
         }
 
-        g_source_attach ((GSource *) client->priv->request_socket,
-                         client->priv->main_context);
-        g_source_unref ((GSource *) client->priv->request_socket);
+        gssdp_socket_source_attach (client->priv->request_socket,
+                                    client->priv->main_context);
 
-        g_source_attach ((GSource *) client->priv->multicast_socket,
-                         client->priv->main_context);
-        g_source_unref ((GSource *) client->priv->multicast_socket);
+        gssdp_socket_source_attach (client->priv->multicast_socket,
+                                    client->priv->main_context);
 }
 
 static void
@@ -234,7 +264,7 @@ gssdp_client_set_property (GObject      *object,
                 client->priv->error = g_value_get_pointer (value);
                 break;
         case PROP_IFACE:
-                client->priv->interface = g_value_dup_string (value);
+                client->priv->iface = g_value_dup_string (value);
                 break;
         case PROP_ACTIVE:
                 client->priv->active = g_value_get_boolean (value);
@@ -254,12 +284,12 @@ gssdp_client_dispose (GObject *object)
 
         /* Destroy the SocketSources */
         if (client->priv->request_socket) {
-                g_source_destroy ((GSource *) client->priv->request_socket);
+                g_object_unref (client->priv->request_socket);
                 client->priv->request_socket = NULL;
         }
 
         if (client->priv->multicast_socket) {
-                g_source_destroy ((GSource *) client->priv->multicast_socket);
+                g_object_unref (client->priv->multicast_socket);
                 client->priv->multicast_socket = NULL;
         }
 
@@ -278,9 +308,12 @@ gssdp_client_finalize (GObject *object)
         GSSDPClient *client;
 
         client = GSSDP_CLIENT (object);
+#ifdef G_OS_WIN32
+        WSACleanup ();
+#endif
 
         g_free (client->priv->server_id);
-        g_free (client->priv->interface);
+        g_free (client->priv->iface);
         g_free (client->priv->host_ip);
 
         G_OBJECT_CLASS (gssdp_client_parent_class)->finalize (object);
@@ -444,12 +477,12 @@ gssdp_client_class_init (GSSDPClientClass *klass)
  **/
 GSSDPClient *
 gssdp_client_new (GMainContext *main_context,
-                  const char   *interface,
+                  const char   *iface,
                   GError      **error)
 {
         return g_object_new (GSSDP_TYPE_CLIENT,
                              "main-context", main_context,
-                             "interface", interface,
+                             "interface", iface,
                              "error", error,
                              NULL);
 }
@@ -535,7 +568,7 @@ gssdp_client_get_interface (GSSDPClient *client)
 {
         g_return_val_if_fail (GSSDP_IS_CLIENT (client), NULL);
 
-        return client->priv->interface;
+        return client->priv->iface;
 }
 
 /**
@@ -583,8 +616,11 @@ _gssdp_client_send_message (GSSDPClient *client,
                             gushort      dest_port,
                             const char  *message)
 {
-        struct sockaddr_in addr;
-        int socket_fd, res;
+        gssize res;
+        GError *error = NULL;
+        GSocket *socket = NULL;
+        GInetAddress *inet_address = NULL;
+        GSocketAddress *address = NULL;
 
         g_return_if_fail (GSSDP_IS_CLIENT (client));
         g_return_if_fail (message != NULL);
@@ -601,25 +637,25 @@ _gssdp_client_send_message (GSSDPClient *client,
         if (dest_port == 0)
                 dest_port = SSDP_PORT;
 
-        socket_fd = gssdp_socket_source_get_fd (client->priv->request_socket);
-
-        memset (&addr, 0, sizeof (addr));
-
-        addr.sin_family      = AF_INET;
-        addr.sin_port        = htons (dest_port);
-        addr.sin_addr.s_addr = inet_addr (dest_ip);
-
-        res = sendto (socket_fd,
-                      message,
-                      strlen (message),
-                      0,
-                      (struct sockaddr *) &addr,
-                      sizeof (addr));
+        inet_address = g_inet_address_new_from_string (dest_ip);
+        address = g_inet_socket_address_new (inet_address, dest_port);
+        socket = gssdp_socket_source_get_socket (client->priv->request_socket);
+        res = g_socket_send_to (socket,
+                                address,
+                                message,
+                                strlen (message),
+                                NULL,
+                                &error);
 
         if (res == -1) {
-                g_warning ("sendto: Error %d sending message: %s",
-                           errno, strerror (errno));
+                g_warning ("Error sending SSDP packet to %s: %s",
+                           dest_ip,
+                           error->message);
+                g_error_free (error);
         }
+
+        g_object_unref (address);
+        g_object_unref (inet_address);
 }
 
 /**
@@ -628,6 +664,19 @@ _gssdp_client_send_message (GSSDPClient *client,
 static char *
 make_server_id (void)
 {
+#ifdef G_OS_WIN32
+        OSVERSIONINFO versioninfo;
+        versioninfo.dwOSVersionInfoSize = sizeof (OSVERSIONINFO);
+        if (GetVersionEx (&versioninfo)) {
+                return g_strdup_printf ("Microsoft Windows/%ld.%ld GSSDP/%s",
+                                        versioninfo.dwMajorVersion,
+                                        versioninfo.dwMinorVersion,
+                                        VERSION);
+        } else {
+                return g_strdup_printf ("Microsoft Windows GSSDP/%s",
+                                        VERSION);
+        }
+#else
         struct utsname sysinfo;
 
         uname (&sysinfo);
@@ -636,6 +685,7 @@ make_server_id (void)
                                 sysinfo.sysname,
                                 sysinfo.version,
                                 VERSION);
+#endif
 }
 
 static gboolean
@@ -706,38 +756,53 @@ parse_http_response (char                *buf,
         }
 }
 
+#ifdef G_OS_WIN32
+static in_addr_t
+inet_netof (struct in_addr in) {
+        in_addr_t i = ntohl(in.s_addr);
+
+        if (IN_CLASSA (i))
+            return (((i) & IN_CLASSA_NET) >> IN_CLASSA_NSHIFT);
+        else if (IN_CLASSB (i))
+            return (((i) & IN_CLASSB_NET) >> IN_CLASSB_NSHIFT);
+        else
+            return (((i) & IN_CLASSC_NET) >> IN_CLASSC_NSHIFT);
+}
+#endif
+
 /**
  * Called when data can be read from the socket
  **/
 static gboolean
-socket_source_cb (GSSDPSocketSource *socket, GSSDPClient *client)
+socket_source_cb (GSSDPSocketSource *socket_source, GSSDPClient *client)
 {
-        int fd, type, len;
-        ssize_t bytes;
+        int type, len;
         char buf[BUF_SIZE], *end;
-        struct sockaddr_in addr;
-        socklen_t addr_size;
         SoupMessageHeaders *headers;
-        struct in_addr our_addr;
-        in_addr_t our_network;
+        GSocket *socket;
+        GSocketAddress *address = NULL;
+        gssize bytes;
+        GInetAddress *inetaddr;
+        char *ip_string;
+        guint16 port;
+        GError *error = NULL;
         in_addr_t recv_network;
+        in_addr_t our_network;
+        struct in_addr our_addr;
+        struct sockaddr_in addr;
 
-        /* Get FD */
-        fd = gssdp_socket_source_get_fd (socket);
-
-        /* Read data */
-        addr_size = sizeof (addr);
-        
-        bytes = recvfrom (fd,
-                          buf,
-                          BUF_SIZE - 1, /* Leave space for trailing \0 */
-                          MSG_TRUNC,
-                          (struct sockaddr *) &addr,
-                          &addr_size);
+        /* Get Socket */
+        socket = gssdp_socket_source_get_socket (socket_source);
+        bytes = g_socket_receive_from (socket,
+                                       &address,
+                                       buf,
+                                       BUF_SIZE - 1,
+                                       NULL,
+                                       &error);
         if (bytes == -1) {
-                g_warning ("Failed to read from socket: %d (%s)",
-                           errno,
-                           strerror (errno));
+                g_warning ("Failed to receive from socket: %s",
+                           error->message);
+                g_error_free (error);
 
                 return TRUE;
         }
@@ -748,6 +813,18 @@ socket_source_cb (GSSDPSocketSource *socket, GSSDPClient *client)
          * on this socket from a particular interface but AFAIK that is not
          * possible, at least not in a portable way.
          */
+
+        if (!g_socket_address_to_native (address,
+                                         &addr,
+                                         sizeof (struct sockaddr_in),
+                                         &error)) {
+                g_warning ("Could not convert address to native: %s",
+                           error->message);
+                g_error_free (error);
+
+                return TRUE;
+        }
+
         recv_network = inet_netof (addr.sin_addr);
         our_addr.s_addr = inet_addr (gssdp_client_get_host_ip (client));
         our_network = inet_netof (our_addr);
@@ -755,6 +832,7 @@ socket_source_cb (GSSDPSocketSource *socket, GSSDPClient *client)
                 return TRUE;
 
         if (bytes >= BUF_SIZE) {
+                g_object_unref (address);
                 g_warning ("Received packet of %u bytes, but the maximum "
                            "buffer size is %d. Packed dropped.",
                            (unsigned int) bytes, BUF_SIZE);
@@ -768,6 +846,7 @@ socket_source_cb (GSSDPSocketSource *socket, GSSDPClient *client)
         /* Find length */
         end = strstr (buf, "\r\n\r\n");
         if (!end) {
+                g_object_unref (address);
                 g_warning ("Received packet lacks \"\\r\\n\\r\\n\" sequence. "
                            "Packed dropped.");
 
@@ -793,24 +872,36 @@ socket_source_cb (GSSDPSocketSource *socket, GSSDPClient *client)
         }
         
         /* Emit signal if parsing succeeded */
+        inetaddr = g_inet_socket_address_get_address (
+                                        G_INET_SOCKET_ADDRESS (address));
+        ip_string = g_inet_address_to_string (inetaddr);
+        port = g_inet_socket_address_get_port (
+                                        G_INET_SOCKET_ADDRESS (address));
         if (type >= 0) {
                 g_signal_emit (client,
                                signals[MESSAGE_RECEIVED],
                                0,
-                               inet_ntoa (addr.sin_addr),
-                               ntohs (addr.sin_port),
+                               ip_string,
+                               port,
                                type,
                                headers);
         }
 
+        if (ip_string)
+                g_free (ip_string);
+
         if (headers)
                 soup_message_headers_free (headers);
 
+        g_object_unref (address);
+
         return TRUE;
 }
 
 static gboolean
-request_socket_source_cb (gpointer user_data)
+request_socket_source_cb (GIOChannel  *source,
+                          GIOCondition condition,
+                          gpointer     user_data)
 {
         GSSDPClient *client;
 
@@ -820,7 +911,9 @@ request_socket_source_cb (gpointer user_data)
 }
 
 static gboolean
-multicast_socket_source_cb (gpointer user_data)
+multicast_socket_source_cb (GIOChannel  *source,
+                            GIOCondition condition,
+                            gpointer     user_data)
 {
         GSSDPClient *client;
 
@@ -829,17 +922,122 @@ multicast_socket_source_cb (gpointer user_data)
         return socket_source_cb (client->priv->multicast_socket, client);
 }
 
+#ifdef G_OS_WIN32
+static gboolean
+is_primary_adapter (PIP_ADAPTER_ADDRESSES adapter)
+{
+        int family =
+                adapter->FirstUnicastAddress->Address.lpSockaddr->sa_family;
+
+        return !(adapter->IfType == IF_TYPE_SOFTWARE_LOOPBACK ||
+                 family == AF_INET6);
+}
+#endif
+
 /*
  * Get the host IP for the specified interface. If no interface is specified,
  * it gets the IP of the first up & running interface and sets @interface
  * appropriately.
  */
 static char *
-get_host_ip (char **interface)
+get_host_ip (char **iface)
 {
+#ifdef G_OS_WIN32
+        char *addr = NULL;
+        GList *up_ifaces = NULL, *ifaceptr = NULL;
+        ULONG flags = GAA_FLAG_INCLUDE_PREFIX |
+                      GAA_FLAG_SKIP_DNS_SERVER |
+                      GAA_FLAG_SKIP_MULTICAST;
+        DWORD size = 15360; /* Use 15k buffer initially as documented in MSDN */
+        DWORD ret;
+        PIP_ADAPTER_ADDRESSES adapters_addresses;
+        PIP_ADAPTER_ADDRESSES adapter;
+
+        do {
+                adapters_addresses = (PIP_ADAPTER_ADDRESSES) g_malloc0 (size);
+                ret = GetAdaptersAddresses (AF_UNSPEC,
+                                            flags,
+                                            NULL,
+                                            adapters_addresses,
+                                            &size);
+                if (ret == ERROR_BUFFER_OVERFLOW)
+                        g_free (adapters_addresses);
+        } while (ret == ERROR_BUFFER_OVERFLOW);
+
+        if (ret == ERROR_SUCCESS)
+                for (adapter = adapters_addresses;
+                     adapter != NULL;
+                     adapter = adapter->Next) {
+                        if (adapter->FirstUnicastAddress == NULL)
+                                continue;
+                        if (adapter->OperStatus != IfOperStatusUp)
+                                continue;
+                        /* skip Point-to-Point devices */
+                        if (adapter->IfType == IF_TYPE_PPP)
+                                continue;
+
+                        if (*iface != NULL &&
+                            strcmp (*iface, adapter->AdapterName) != 0)
+                                continue;
+
+                        /* I think that IPv6 is done via pseudo-adapters, so
+                         * that there are either IPv4 or IPv6 addresses defined
+                         * on the adapter.
+                         * Loopback-Devices and IPv6 go to the end of the list,
+                         * IPv4 to the front
+                         */
+                        if (is_primary_adapter (adapter))
+                                up_ifaces = g_list_prepend (up_ifaces, adapter);
+                        else
+                                up_ifaces = g_list_append (up_ifaces, adapter);
+                }
+
+        for (ifaceptr = up_ifaces;
+             ifaceptr != NULL;
+             ifaceptr = ifaceptr->next) {
+                char ip[INET6_ADDRSTRLEN];
+                DWORD len = INET6_ADDRSTRLEN;
+                const char *p = NULL;
+                PIP_ADAPTER_ADDRESSES adapter;
+                SOCKET_ADDRESS unicast_address;
+
+                adapter = (PIP_ADAPTER_ADDRESSES) ifaceptr->data;
+                unicast_address = adapter->FirstUnicastAddress->Address;
+
+                switch (unicast_address.lpSockaddr->sa_family) {
+                        case AF_INET:
+                        case AF_INET6:
+                                ret = WSAAddressToStringA (
+                                        unicast_address.lpSockaddr,
+                                        unicast_address.iSockaddrLength,
+                                        NULL,
+                                        ip,
+                                        &len);
+                                if (ret == 0)
+                                        p = ip;
+
+                                break;
+                        default:
+                                continue;
+                }
+
+                if (p != NULL) {
+                        addr = g_strdup (p);
+                        if (*iface == NULL)
+                                *iface = g_strdup (adapter->AdapterName);
+
+                        break;
+                }
+
+        }
+        g_list_free (up_ifaces);
+        g_free (adapters_addresses);
+
+        return addr;
+#else
         struct ifaddrs *ifa_list, *ifa;
         char *ret;
-        GList *up_ifaces, *iface;
+        GList *up_ifaces, *ifaceptr;
 
         ret = NULL;
         up_ifaces = NULL;
@@ -855,7 +1053,7 @@ get_host_ip (char **interface)
                 if (ifa->ifa_addr == NULL)
                         continue;
 
-                if (*interface && strcmp (*interface, ifa->ifa_name) != 0)
+                if (*iface && strcmp (*iface, ifa->ifa_name) != 0)
                         continue;
                 else if (!(ifa->ifa_flags & IFF_UP))
                         continue;
@@ -870,7 +1068,9 @@ get_host_ip (char **interface)
                         up_ifaces = g_list_prepend (up_ifaces, ifa);
         }
 
-        for (iface = up_ifaces; iface != NULL; iface = iface->next) {
+        for (ifaceptr = up_ifaces;
+             ifaceptr != NULL;
+             ifaceptr = ifaceptr->next) {
                 char ip[INET6_ADDRSTRLEN];
                 const char *p;
                 struct sockaddr_in *s4;
@@ -878,7 +1078,7 @@ get_host_ip (char **interface)
 
                 p = NULL;
 
-                ifa = iface->data;
+                ifa = ifaceptr->data;
 
                 switch (ifa->ifa_addr->sa_family) {
                 case AF_INET:
@@ -898,8 +1098,8 @@ get_host_ip (char **interface)
                 if (p != NULL) {
                         ret = g_strdup (p);
 
-                        if (*interface == NULL)
-                                *interface = g_strdup (ifa->ifa_name);
+                        if (*iface == NULL)
+                                *iface = g_strdup (ifa->ifa_name);
                         break;
                 }
         }
@@ -908,6 +1108,7 @@ get_host_ip (char **interface)
         freeifaddrs (ifa_list);
 
         return ret;
+#endif
 }
 
 static gboolean
@@ -915,11 +1116,11 @@ init_network_info (GSSDPClient *client)
 {
         gboolean ret = TRUE;
 
-        if (client->priv->interface == NULL || client->priv->host_ip == NULL)
+        if (client->priv->iface == NULL || client->priv->host_ip == NULL)
                 client->priv->host_ip =
-                        get_host_ip (&client->priv->interface);
+                        get_host_ip (&client->priv->iface);
 
-        if (client->priv->interface == NULL) {
+        if (client->priv->iface == NULL) {
                 if (client->priv->error)
                         g_set_error (client->priv->error,
                                      GSSDP_ERROR,
@@ -933,7 +1134,7 @@ init_network_info (GSSDPClient *client)
                                      GSSDP_ERROR,
                                      GSSDP_ERROR_NO_IP_ADDRESS,
                                      "Failed to find IP of interface %s",
-                                     client->priv->interface);
+                                     client->priv->iface);
 
                 ret = FALSE;
         }
